"""Package teneva, module core.als: construct TT-tensor, using TT-ALS.

This module contains the function "als" which computes the TT-approximation
for the tensor by TT-ALS algorithm, using given random samples.

"""
import numpy as np
import scipy as sp


from .tensor import copy
from .tensor import get
from .transformation import orthogonalize
from .transformation import orthogonalize_left
from .transformation import orthogonalize_right


def als(I_trn, Y_trn, Y0, nswp=10):
    """Build TT-tensor by TT-ALS from the given random tensor samples.

    Args:
        I_trn (np.ndarray): multi-indices for the tensor in the form of array
            of the shape [samples, d].
        Y_trn (np.ndarray): values of the tensor for multi-indices I in the form
            of array of the shape [samples].
        Y0 (list): TT-tensor, which is the initial approximation for algorithm.
        nswp (int): number of ALS iterations (sweeps).

    Returns:
        list: TT-tensor, which represents the TT-approximation for the tensor.

    """
    I_trn = np.asanyarray(I_trn, dtype=int)
    Y_trn = np.asanyarray(Y_trn, dtype=float)

    m = I_trn.shape[0]
    d = I_trn.shape[1]
    Y = copy(Y0)

    for k in range(d):
        if np.unique(I_trn[:, k]).size != Y[k].shape[1]:
            raise ValueError('One groundtruth sample is needed for every slice')

    Yl = [np.ones((1, m, Y[k].shape[0])) for k in range(d)]
    Yr = [None for _ in range(d-1)] + [np.ones((1, m, 1))]

    orthogonalize(Y, 0)

    for k in range(d-1, 0, -1):
        i_trn = I_trn[:, k]
        Yr[k-1] = np.einsum('ijk,kjl->ijl', Y[k][:, i_trn, :], Yr[k])

    for _ in range(nswp):
        for k in range(d-1):
            i_trn = I_trn[:, k]
            optimize_core(Y[k], i_trn, Y_trn, Yl[k], Yr[k])
            orthogonalize_left(Y, k)
            Yl[k+1] = np.einsum('ijk,kjl->ijl', Yl[k], Y[k][:, i_trn, :])

        for k in range(d-1, 0, -1):
            i_trn = I_trn[:, k]
            optimize_core(Y[k], i_trn, Y_trn, Yl[k], Yr[k])
            orthogonalize_right(Y, k)
            Yr[k-1] = np.einsum('ijk,kjl->ijl', Y[k][:, i_trn, :], Yr[k])

    return Y


def als2(I_trn, Y_trn, Y0, nswp=10, eps=None):
    """Build TT-tensor by TT-ALS from the given random tensor samples.

    Args:
        I_trn (np.ndarray): multi-indices for the tensor in the form of array
            of the shape [samples, d].
        Y_trn (np.ndarray): values of the tensor for multi-indices I in the form
            of array of the shape [samples].
        Y0 (list): TT-tensor, which is the initial approximation for algorithm.
        nswp (int):  number of ALS iterations (sweeps).
        eps (float): desired accuracy of approximation.

    Returns:
        list: TT-tensor, which represents the TT-approximation for the tensor.

    Note:
        This is the alternative realization of the algorithm. The version from
        "als" function in many cases works better and much faster.

        Note that the code of this function is not optimized.

    """
    I_trn = np.asanyarray(I_trn, dtype=int)
    Y_trn = np.asanyarray(Y_trn, dtype=float)

    P, d = I_trn.shape

    norm = np.linalg.norm(Y_trn)
    Y_trn = Y_trn.copy()
    Y_trn /= norm

    Y = [G.copy() for G in Y0]

    elist = []

    for swp in range(nswp):

        for k in range(d):
            r1, n, r2 = Y[k].shape

            core = np.zeros([r1, n, r2])

            leftU = Y[:k] if k > 0 else []
            rightV = Y[k+1:] if k < d-1 else []

            for i in range(n):
                thetaI = np.where(I_trn[:, k] == i)[0]
                if len(thetaI) == 0:
                    continue

                A = np.zeros([len(thetaI), r1*r2])
                for j in range(len(thetaI)):
                    A[j:j+1, :] += getRow(leftU, rightV, I_trn[thetaI[j], :])

                vec_slice = np.linalg.lstsq(A, Y_trn[thetaI], rcond=-1)[:1]
                core[:, i, :] += _reshape(vec_slice, [r1, r2])

            Y[k] = core.copy()

        if eps is not None:
            get = lambda x: get(Y, x)
            e = 0.5 * sum((get(I_trn[p, :]) - Y_trn[p])**2 for p in range(P))
            elist.append(e)
            if e < eps or swp > 0 and abs(e - elist[-2]) < eps:
                break

    Y[0] *= norm

    return Y


def getRow(leftU, rightV, jVec):
    jLeft = jVec[:len(leftU)] if len(leftU) > 0 else None
    jRight = jVec[-len(rightV):] if len(rightV) > 0 else None

    multU = np.ones([1, 1])
    for k in range(len(leftU)):
        multU = multU @ leftU[k][:, jLeft[k], :]

    multV= np.ones([1, 1])
    for k in range(len(rightV)-1, -1, -1):
        multV = rightV[k][:, jRight[k], :] @ multV

    return np.kron(multV.T, multU)


def optimize_core(G, i_trn, Y_trn, left, right):
    for k in range(G.shape[1]):
        idx = np.where(i_trn == k)[0]

        leftside = left[0, idx, :]
        rightside = right[:, idx, 0]

        lhs = np.transpose(rightside, [1, 0])[:, :, np.newaxis]
        rhs = leftside[:, np.newaxis, :]
        A = _reshape(lhs * rhs, (len(idx), -1))

        b = Y_trn[idx]

        sol, residuals = sp.linalg.lstsq(A, b)[0:2]
        if residuals.size == 0:
            residuals = np.linalg.norm(A.dot(sol) - b) ** 2

        G[:, k, :] = _reshape(sol, G[:, k, :].shape, 'C')


def _reshape(A, n, order='F'):
    return np.reshape(A, n, order=order)



"""Package teneva, module func.demo.func_demo_michalewicz: function.

This module contains class that implements analytical Michalewicz function
for demo and tests.

"""
import numpy as np


from ..func import Func


class FuncDemoMichalewicz(Func):
    def __init__(self, d, m=10.):
        """Michalewicz function for demo and tests.

        Args:
            d (int): number of dimensions.
            m (float): parameter of the function.

        Note:
            See https://www.sfu.ca/~ssurjano/michal.html for details.

            See also Charlie Vanaret, Jean-Baptiste Gotteland, Nicolas Durand,
            Jean-Marc Alliot. "Certified global minima for a benchmark of
            difficult optimization problems". arXiv preprint arXiv:2003.09867
            2020.

            Note that the value of the global minimum is known only for the
            case of dimensions 2, 5, and 10. In this cases, only the
            corresponding value of the function is known, but not the argument.

        """
        super().__init__(d, name='Michalewicz')

        self.par_m = m

        self.set_lim(0., np.pi)

        if self.d == 2:
            self.set_min([2.20, 1.57], -1.8013)
        if self.d == 5:
            self.set_min(None, -4.687658)
        if self.d == 10:
            self.set_min(None, -9.66015)

    def _calc(self, x):
        y1 = np.sin(((np.arange(self.d) + 1) * x**2 / np.pi))

        y = -np.sum(np.sin(x) * y1**(2 * self.par_m))

        return y

    def _comp(self, X):
        y1 = np.sin(((np.arange(self.d) + 1) * X**2 / np.pi))

        y = -np.sum(np.sin(X) * y1**(2 * self.par_m), axis=1)

        return y



from .func_demo_ackley import FuncDemoAckley
from .func_demo_brown import FuncDemoBrown
from .func_demo_grienwank import FuncDemoGrienwank
from .func_demo_michalewicz import FuncDemoMichalewicz
from .func_demo_piston import FuncDemoPiston
from .func_demo_rastrigin import FuncDemoRastrigin
from .func_demo_rosenbrock import FuncDemoRosenbrock
from .func_demo_schaffer import FuncDemoSchaffer
from .func_demo_schwefel import FuncDemoSchwefel


def func_demo_all(d, names=None, with_piston=False):
    """Build list of class instances for all demo functions.

    Args:
        d (int): number of dimensions.
        names (list): optional list of function names (in any register),
            which should be added to the resulting list of class instances.
            The following functions are available: "ackley", "brown",
            "grienwank", "michalewicz", "piston", "rastrigin", "rosenbrock",
            "schaffer" and "schwefel".
        with_piston (bool): If True, then Piston function will be also
            added to the list. Note that this function is 7-dimensional,
            hence given argument "d" will be ignored for this function. The
            value of this flag does not matter if the names ("names" argument)
            of the added functions are explicitly specified.

    Returns:
        list: the list of class instances for all demo functions.

    """
    funcs_all = [
        FuncDemoAckley(d),
        FuncDemoBrown(d),
        FuncDemoGrienwank(d),
        FuncDemoMichalewicz(d),
        FuncDemoPiston(d=7),
        FuncDemoRastrigin(d),
        FuncDemoRosenbrock(d),
        FuncDemoSchaffer(d),
        FuncDemoSchwefel(d),
    ]

    if names is not None:
        names = [name.lower() for name in names]

    funcs = []
    for func in funcs_all:
        if names is not None:
            if func.name.lower() not in names:
                continue
        else:
            if not with_piston and func.name == 'Piston':
                continue

        funcs.append(func)

    return funcs



"""Package teneva, module core.transformation: transformation of TT-tensors.

This module contains the functions for orthogonalization and truncation of the
TT-tensors.

"""
import numpy as np
import scipy as sp


from .svd import matrix_svd


def orthogonalize(Y, k=None):
    """Orthogonalize TT-tensor (note that operation is performed inplace).

    Args:
        Y (list): TT-tensor.
        k (int): the leading mode for orthogonalization. The TT-cores 0, 1, ...,
            k-1 will be left-orthogonalized and the TT-cores k+1, k+2, ..., d-1
            will be right-orthogonalized. It will be the last mode by default.

    Returns:
        [np.ndarray, np.ndarray]: R and L factor matrices.

    """
    d = len(Y)
    k = d - 1 if k is None else k
    if k is None or k < 0 or k > d-1:
        raise ValueError('Invalid mode number')

    R = np.array([[1.]])
    for i in range(k):
        R = orthogonalize_left(Y, i)

    L = np.array([[1.]])
    for i in range(d-1, k, -1):
        L = orthogonalize_right(Y, i)

    return R, L


def orthogonalize_left(Y, k):
    """Left-orthogonalization for TT-tensor (operation is performed inplace).

    Args:
        Y (list): d-dimensional TT-tensor.
        k (int): mode for orthogonalization (>= 0 and < d-1).

    Returns:
        np.ndarray: R factor matrix.

    """
    d = len(Y)
    if k is None or k < 0 or k >= d-1:
        raise ValueError('Invalid mode number')

    r1, n1, r2 = Y[k].shape
    G1 = _reshape(Y[k], (r1 * n1, r2))
    Q, R = np.linalg.qr(G1, mode='reduced')
    Y[k] = _reshape(Q, (r1, n1, Q.shape[1]))

    r2, n2, r3 = Y[k+1].shape
    G2 = _reshape(Y[k+1], (r2, n2 * r3))
    G2 = R @ G2
    Y[k+1] = _reshape(G2, (G2.shape[0], n2, r3))

    return R

def orthogonalize_right(Y, k):
    """Right-orthogonalization for TT-tensor (operation is performed inplace).

    Args:
        Y (list): d-dimensional TT-tensor.
        k (int): mode for orthogonalization (> 0 and <= d-1).

    Returns:
        np.ndarray: L factor matrix.

    """
    d = len(Y)
    if k is None or k <= 0 or k > d-1:
        raise ValueError('Invalid mode number')

    r2, n2, r3 = Y[k].shape
    G2 = _reshape(Y[k], (r2, n2 * r3))
    L, Q = sp.linalg.rq(G2, mode='economic', check_finite=False)
    Y[k] = _reshape(Q, (Q.shape[0], n2, r3))

    r1, n1, r2 = Y[k-1].shape
    G1 = _reshape(Y[k-1], (r1 * n1, r2))
    G1 = G1 @ L
    Y[k-1] = _reshape(G1, (r1, n1, G1.shape[1]))

    return L


def truncate(Y, e, r=1.E+12, orth=True):
    """Truncate (round) TT-tensor.

    Args:
        Y (list): TT-tensor wth overestimated ranks.
        e (float): desired approximation accuracy (> 0).
        r (int, float): maximum TT-rank of the result (> 0).
        orth (bool): if the flag is set, then tensor orthogonalization will be
            performed (it is True by default).

    Returns:
        list: TT-tensor, which is rounded up to a given accuracy "e" and
        satisfying the rank constraint "r".

    """
    d = len(Y)
    Z = [G.copy() for G in Y]

    if orth:
        orthogonalize(Z)
        e = e / np.sqrt(d-1) * np.linalg.norm(Z[-1])

    for k in range(d-1, 0, -1):
        r1, n, r2 = Z[k].shape
        G = _reshape(Z[k], (r1, n * r2))
        U, V = matrix_svd(G, e, r)
        Z[k] = _reshape(V, (-1, n, r2))
        Z[k-1] = np.einsum('ijq,ql', Z[k-1], U, optimize=True)

    return Z


def _reshape(A, n):
    return np.reshape(A, n, order='F')



"""Package teneva, module core.stat: helper functions for processing statistics.

This module contains the helper functions for processing statistics, including
computation of the CDF function and its confidence bounds.

"""
import numpy as np


def cdf_confidence(x, alpha=0.05):
    """Construct a Dvoretzky-Kiefer-Wolfowitz confidence band for the CDF.

    Args:
        x (np.ndarray): the empirical distribution in the form of 1D np.ndarray
            of length "m".
        alpha (float): "alpha" for the "(1 - alpha)" confidence band.

    Returns:
        [np.ndarray, np.ndarray]: CDF lower and upper bounds in the form of 1D
        np.ndarray of the length "m".

    """
    eps = np.sqrt(np.log(2. / alpha) / (2 * len(x)))
    return np.clip(x - eps, 0, 1), np.clip(x + eps, 0, 1)


def cdf_getter(x):
    """Build the getter for CDF.

    Args:
        x (list or np.ndarray): one-dimensional points.

    Returns:
        function: the function that computes CDF values. Its input may be one
        point (float) or a set of points (1D np.ndarray). The output
        (corresponding CDF value/values) will have the same type.

    """
    x = np.array(x, copy=True)
    x.sort()
    y = np.linspace(1./len(x), 1, len(x))

    x = np.r_[-np.inf, x]
    y = np.r_[0, y]

    def cdf(z):
        return y[np.searchsorted(x, z, 'right') - 1]

    return cdf



"""Package teneva, module core.cross: cross approximation in the TT-format.

This module contains the function "cross" which computes the TT-approximation
for implicit tensor given functionally by multidimensional cross approximation
method in the TT-format (TT-CROSS).

"""
import numpy as np


from .grid import ind_to_str
from .maxvol import maxvol
from .maxvol import maxvol_rect
from .tensor import accuracy
from .tensor import copy
from .tensor import shape
from .transformation import truncate


def cross(f, Y0, e, m=None, nswp=10, dr_min=1, dr_max=2, info={}, cache=None):
    """Compute the TT-approximation for implicit tensor given functionally.

    Args:
        f (function): function f(I) which computes tensor elements for the given
            set of multi-indices I, where I is a 2D np.ndarray of the shape
            [samples, dimensions]. The function should return 1D np.ndarray of
            the length equals to "samples".
        Y0 (list): TT-tensor, which is the initial approximation for algorithm.
        e (float): accuracy (> 0) for truncation of the final result and
            algorithm convergence criterion. If between iterations the relative
            rate of solution change is less than this value, then the operation
            of the algorithm will be interrupted.
        m (int): an optionally set limit on the maximum number of requests to
            the objective function. If specified, then the total number of
            requests will not exceed this value. Note that the actual number of
            requests may be less, since the values are requested in batches.
        nswp (int): maximum number of iterations (sweeps) of the algorithm.
        dr_min (int): minimum number of added rows in the process of adaptively
            increasing the TT-rank of the approximation using the algorithm
            maxvol_rect (see "maxvol_rect" function for more details). Note
            that "dr_min" should be no bigger than "dr_max".
        dr_max (int): maximum number of added rows in the process of adaptively
            increasing the TT-rank of the approximation using the algorithm
            maxvol_rect (see "maxvol_rect" function for more details). Note
            that "dr_max" should be no less than "dr_min".
        info (dict): an optionally set dictionary, which will be filled with
            reference information about the process of the algorithm operation.
            At the end of the function work, it will contain parameters: "e" -
            the final value of the convergence criterion; "m" - total number of
            requests to target function "f"; "m_cache" - total number of
            requests to cache; "nswp" - the real number of performed iterations
            (sweeps); "stop" - stop type of the algorithm (see note below).
        cache (dict): an optionally set dictionary, which will be filled with
            requested function values. Since the algorithm sometimes requests
            the same tensor indices, the use of such a cache may speed up the
            operation of the algorithm if the time to find a value in the cache
            is less than the time to calculate the function.

    Returns:
        list: TT-Tensor which approximates the implicit tensor.

    Note:
        Note that the end of the algorithm operation occurs when one of the
        following criteria is reached: 1) the maximum number of iterations
        ("nswp") performed; 2) the maximum allowable number of the objective
        function calls ("m") has been done (more precisely, if the next request
        will result in exceeding this value, then algorithm will not perform
        this new request). If "m" is set, then other criteria will be ignored;
        3) the convergence criterion ("e") is reached; 4) the algorithm is
        already converged (all requested values are in the cache already). The
        related stop type ("nswp", "m", "e", "conv") will be written into the
        item "stop" of the "info" dictionary.

    """
    info['e'] = -1.
    info['m'] = 0
    info['m_cache'] = 0
    info['m_max'] = int(m) if m else None
    info['nswp'] = 0
    info['stop'] = None

    Y = copy(Y0)
    d = len(Y)
    n = shape(Y)

    Ig = [_reshape(np.arange(k, dtype=int), (-1, 1)) for k in n] # Grid
    Ir = [None for i in range(d+1)]                              # Row indices
    Ic = [None for i in range(d+1)]                              # Col indices

    R = np.ones((1, 1))
    for i in range(d):
        G = np.tensordot(R, Y[i], 1)
        Y[i], R, Ir[i+1] = _iter(G, Ig[i], Ir[i], l2r=True)
    Y[d-1] = np.tensordot(Y[d-1], R, 1)

    R = np.ones((1, 1))
    for i in range(d-1, -1, -1):
        G = np.tensordot(Y[i], R, 1)
        Y[i], R, Ic[i] = _iter(G, Ig[i], Ic[i+1], l2r=False)
    Y[0] = np.tensordot(R, Y[0], 1)

    Yold = None
    while True:
        R = np.ones((1, 1))
        for i in range(d):
            Z = _func(f, Ig[i], Ir[i], Ic[i+1], info, cache)
            if Z is None:
                info['stop'] = 'm'
                Y[i] = np.tensordot(R, Y[i], 1)
                return truncate(Y, e)
            Y[i], R, Ir[i+1] = _iter(Z, Ig[i], Ir[i], dr_min, dr_max, l2r=True)
        Y[d-1] = np.tensordot(Y[d-1], R, 1)

        R = np.ones((1, 1))
        for i in range(d-1, -1, -1):
            Z = _func(f, Ig[i], Ir[i], Ic[i+1], info, cache)
            if Z is None:
                info['stop'] = 'm'
                Y[i] = np.tensordot(Y[i], R, 1)
                return truncate(Y, e)
            Y[i], R, Ic[i] = _iter(Z, Ig[i], Ic[i+1], dr_min, dr_max, l2r=False)
        Y[0] = np.tensordot(R, Y[0], 1)

        if Yold is not None:
            eps = accuracy(Y, Yold)
            info['e'] = eps

            if not info['m_max'] and eps <= e:
                info['stop'] = 'e'
                return truncate(Y, e)

        info['nswp'] += 1
        if not info['m_max'] and info['nswp'] >= nswp:
            info['stop'] = 'nswp'
            return truncate(Y, e)

        if info['m_cache'] > 5 * info['m']:
            info['stop'] = 'conv'
            return truncate(Y, e)

        Yold = copy(Y)


def _func(f, Ig, Ir, Ic, info, cache=None):
    n = Ig.shape[0]
    r1 = Ir.shape[0] if Ir is not None else 1
    r2 = Ic.shape[0] if Ic is not None else 1

    I = np.kron(np.kron(_ones(r2), Ig), _ones(r1))
    if Ir is not None:
        Ir_ = np.kron(_ones(n * r2), Ir)
        I = np.hstack((Ir_, I))
    if Ic is not None:
        Ic_ = np.kron(Ic, _ones(r1 * n))
        I = np.hstack((I, Ic_))

    y = _func_eval(f, I, info, cache)
    if y is not None:
        return _reshape(y, (r1, n, r2))


def _func_eval(f, I, info, cache=None):
    if cache is None:
        if info['m_max'] and info['m'] + len(I) > info['m_max']:
            return None
        info['m'] += len(I)
        return f(I)

    I_new = np.array([i for i in I if ind_to_str(i) not in cache])
    if len(I_new):
        if info['m_max'] and info['m'] + len(I_new) > info['m_max']:
            return None
        Y_new = f(I_new)
        for k, i in enumerate(I_new):
            cache[ind_to_str(i)] = Y_new[k]

    info['m'] += len(I_new)
    info['m_cache'] += len(I) - len(I_new)

    return np.array([cache[ind_to_str(i)] for i in I])


def _iter(Z, Ig, I=None, dr_min=0, dr_max=0, l2r=True):
    r1, n, r2 = Z.shape
    r = r1 if l2r else r2
    Z = _reshape(Z, (r1 * n, r2)) if l2r else _reshape(Z, (r1, n * r2)).T

    Q, R = np.linalg.qr(Z)
    ind, B = _maxvol(Q, dr_min, dr_max)

    G = B if l2r else B.T
    G = _reshape(G, (r1, n, -1)) if l2r else _reshape(G, (-1, n, r2))

    R = Q[ind, :] @ R
    R = R if l2r else R.T

    I_new = np.kron(Ig, _ones(r)) if l2r else np.kron(_ones(r), Ig)
    if I is not None:
        I_old = np.kron(_ones(n), I) if l2r else np.kron(I, _ones(n))
        I_new = np.hstack((I_old, I_new)) if l2r else np.hstack((I_new, I_old))
    I_new = I_new[ind, :]

    return G, R, I_new


def _maxvol(A, dr_min=0, dr_max=0, tau=1.1):
    n, r = A.shape
    dr_max = min(dr_max, n - r)
    dr_min = min(dr_min, dr_max)

    if n <= r:
        I = np.arange(n, dtype=int)
        B = np.eye(n, dtype=float)
    elif dr_max == 0:
        I, B = maxvol(A)
    else:
        I, B = maxvol_rect(A, tau, dr_min, dr_max)

    return I, B


def _ones(k, m=1):
    return np.ones((k, m), dtype=int)


def _reshape(A, shape):
    return np.reshape(A, shape, order='F')



"""Package teneva, module func.demo.func_demo_schwefel: function.

This module contains class that implements analytical Schwefel function
for demo and tests.

"""
import numpy as np


from ..func import Func


class FuncDemoSchwefel(Func):
    def __init__(self, d):
        """Schwefel function for demo and tests.

        Args:
            d (int): number of dimensions.

        Note:
            See https://www.sfu.ca/~ssurjano/schwef.html for details.

            See also Johannes M Dieterich, Bernd Hartke. "Empirical review of
            standard benchmark functions using evolutionary global
            optimization". Applied Mathematics 2012; 3:1552-1564.

        """
        super().__init__(d, name='Schwefel')

        self.set_lim(-500., +500.)
        self.set_min([420.9687]*self.d, 0.)

    def _calc(self, x):
        return 418.9829*self.d - np.sum(x * np.sin(np.sqrt(np.abs(x))))

    def _comp(self, X):
        return 418.9829*self.d - np.sum(X * np.sin(np.sqrt(np.abs(X))), axis=1)



"""Package teneva, module func.demo.func_demo_piston: function.

This module contains class that implements analytical Piston function
for demo and tests.

"""
import numpy as np


from ..func import Func


class FuncDemoPiston(Func):
    def __init__(self, d=7):
        """Piston 7-dimensional function for demo and tests.

        Args:
            d (int): number of dimensions. It should be 7.

        Note:
            See Vitaly Zankin, Gleb Ryzhakov, Ivan Oseledets. "Gradient descent
            based D-optimal design for the least-squares polynomial
            approximation". arXiv preprint arXiv:1806.06631 2018 for details.

            Note that the value of the global minimum is unknown.

        """
        super().__init__(d, name='Piston')

        if self.d != 7:
            raise ValueError('DemoFuncPiston is available only for 7-d case')

        self.set_lim(
            [30., 0.005, 0.002, 1000,  90000, 290, 340],
            [60., 0.020, 0.010, 5000, 110000, 296, 360])

    def _calc(self, x):
        return self._comp(x.reshape((1, -1)))[0]

    def _comp(self, X):
        _M  = X[:, 0]
        _S  = X[:, 1]
        _V0 = X[:, 2]
        _k  = X[:, 3]
        _P0 = X[:, 4]
        _Ta = X[:, 5]
        _T0 = X[:, 6]

        _A = _P0 * _S + 19.62 * _M - _k * _V0 / _S
        _Q = _P0 * _V0 / _T0
        _V = _S / 2 / _k * (np.sqrt(_A**2 + 4 * _k * _Q * _Ta) - _A)
        _C = 2 * np.pi * np.sqrt(_M / (_k + _S**2 * _Q * _Ta / _V**2))

        return _C



"""Package teneva, module core.cheb_full: Chebyshev interpolation.

This module contains the functions for construction of the Chebyshev
interpolation in the full format as well as calculating the values of the
function using the constructed interpolation coefficients.

Note:
    See module "core.cheb" with the same functions in the TT-format.

"""
import numpy as np


from .cheb import cheb_pol
from .grid import grid_flat
from .grid import grid_prep_opt
from .grid import grid_prep_opts
from .grid import ind_to_poi


def cheb_bld_full(f, a, b, n, **args):
    """Compute the function values on the Chebyshev grid.

    Args:
        f (function): function f(X) for interpolation, where X should be 2D
            np.ndarray of the shape [samples, dimensions]. The function should
            return 1D np.ndarray of the length equals to "samples".
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the upper
            bounds for each dimension will be the same.
        n (int, float, list, np.ndarray): tensor size for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the size
            for each dimension will be the same.
        args (dict): named arguments, which are not used in this function
            (added for compatibility with "cheb_bld" function).

    Returns:
        np.ndarray: d-dimensional array with function values on the Chebyshev
        grid.

    Note:
        At least one of the variables "a", "b", "n" must be a list or
        np.ndarray (to be able to automatically determine the dimension).

        See also the same function ("cheb_bld") in the TT-format.

    """
    a, b, n = grid_prep_opts(a, b, n)
    I = grid_flat(n)
    X = ind_to_poi(I, a, b, n, 'cheb')
    Y = f(X).reshape(n, order='F')
    return Y


def cheb_get_full(X, A, a, b, z=0.):
    """Compute the Chebyshev approximation in given points (approx. f(X)).

    Args:
        X (np.ndarray): spatial points of interest (it is 2D array of the shape
            [samples, d], where "d" is the number of dimensions).
        A (np.ndarray): d-dimensional tensor of the interpolation coefficients.
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the upper
            bounds for each dimension will be the same.
        z (float): the value for points, which are outside the spatial grid.

    Returns:
        np.ndarray: approximated function values in X points (it is 1D array of
        the shape [samples]).

    Note:
        See also the same function ("cheb_get") in the TT-format.

    """
    d = len(A.shape)
    n = A.shape
    m = X.shape[0]
    a, b, n = grid_prep_opts(a, b, n, d)

    T = cheb_pol(X, a, b, max(n))

    Y = np.ones(m) * z
    for i in range(m):
        if np.max(a - X[i, :]) > 1.E-16 or np.max(X[i, :] - b) > 1.E-16:
            # We skip the points outside the grid bounds:
            continue

        Q = A.copy()
        for j in range(d):
            Q = np.tensordot(Q, T[:n[j], i, j], axes=([0], [0]))
        Y[i] = Q

    return Y


def cheb_gets_full(A, a, b, m=None):
    """Compute the Chebyshev approximation all over the new grid.

    Args:
        A (np.ndarray): d-dimensional tensor of the interpolation coefficients.
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the upper
            bounds for each dimension will be the same.
        m (int, float, list, np.ndarray): tensor size for each dimension of the
            new grid (list or np.ndarray of length "d"). It may be also
            int/float, then the size for each dimension will be the same. If it
            is not set, then original grid size (from the interpolation) will be
            used.

    Returns:
        list: array of the approximated function values on the full new grid.
        This is the d-dimensional array of the shape "m".

    Note:
        This function is not efficient in the full format, and corresponds to a
        simple explicit calculation of all values at the nodes of the new grid.
        At the same time, in the TT-format ("cheb_gets"), the corresponding
        operation is carried out in an implicit efficient form. Accordingly,
        this function is provided only for uniformity of presentation.

    """
    d = len(A.shape)
    n = A.shape
    a, b, n = grid_prep_opts(a, b, n, d)
    m = n if m is None else grid_prep_opt(m, d, int)

    I = grid_flat(m)
    X = ind_to_poi(I, a, b, m, 'cheb')
    Z = cheb_get_full(X, A, a, b)
    Z = Z.reshape(m, order='F')

    return Z


def cheb_int_full(Y):
    """Compute the tensor for Chebyshev interpolation coefficients.

    Args:
        Y (np.ndarray): d-dimensional array with function values on the
            Chebyshev grid.

    Returns:
        np.ndarray: array that collects interpolation coefficients. It has the
        same shape as the given tensor Y.

    Note:
        See also the same function ("cheb_int") in the TT-format.

    """
    d = len(Y.shape)
    A = Y.copy()
    for k in range(d):
        n = np.array(Y.shape, dtype=int)
        m = n[k]
        n[[0, k]] = n[[k, 0]]
        A = np.swapaxes(A, 0, k)
        A = A.reshape((m, -1), order='F')
        A = np.vstack([A, A[m-2 : 0 : -1, :]])
        A = np.fft.fft(A, axis=0).real
        A = A[:m, :] / (m - 1)
        A[0, :] /= 2.
        A[m-1, :] /= 2.
        A = A.reshape(n, order='F')
        A = np.swapaxes(A, 0, k)
    return A


def cheb_sum_full(A, a, b):
    """Integrate the function from its Chebyshev approximation.

    Args:
        A (np.ndarray): d-dimensional tensor of the interpolation coefficients.
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the upper
            bounds for each dimension will be the same.

    Returns:
        float: the value of the integral.

    Note:
        This function works only for symmetric grids!

        See also the same function ("cheb_sum") in the TT-format.

    """
    d = len(A.shape)
    n = A.shape
    a, b, n = grid_prep_opts(a, b, n, d)

    for k in range(d):
        if abs(abs(b[k]) - abs(a[k])) > 1.E-16:
            raise ValueError('This function works only for symmetric grids')

    v = A.copy()
    for k in range(d):
        v = v.reshape(n[k], -1)
        p = np.arange(n[k])[::2]
        p = np.repeat(p.reshape(-1, 1), v.shape[1], axis=1)
        v = np.sum(v[::2, :] * 2. / (1. - p**2), axis=0)
        v*= (b[k] - a[k]) / 2.

    return v[0]



"""Package teneva, module func.demo.func_demo_rosenbrock: function.

This module contains class that implements analytical Rosenbrock function
for demo and tests.

"""
import numpy as np
from scipy.optimize import rosen


from ..func import Func


class FuncDemoRosenbrock(Func):
    def __init__(self, d):
        """Rosenbrock function for demo and tests.

        Args:
            d (int): number of dimensions.

        Note:
            See https://www.sfu.ca/~ssurjano/rosen.html for details.

            See also Momin Jamil, Xin-She Yang. "A literature survey of
            benchmark functions for global optimization problems". Journal of
            Mathematical Modelling and Numerical Optimisation 2013; 4:150-194.

        """
        super().__init__(d, name='Rosenbrock')

        self.set_lim(-2.048, +2.048)
        self.set_min([1.]*self.d, 0.)

    def _calc(self, x):
        return rosen(x)

    def _comp(self, X):
        return rosen(X.T)



"""Package teneva, module core.svd: SVD-based algorithms.

This module contains the basic implementation of the TT-SVD algorithm (function
svd) as well as new original TT-SVD-incomplete algorithm (function
svd_incomplete), which implements efficient construction of the TT-tensor based
on specially selected elements. This module also contains functions for
constructing the SVD decomposition (function matrix_svd) and skeleton
decomposition (matrix_skeleton) for the matrices.

"""
import numpy as np


from .tensor import get


def matrix_skeleton(A, e=1.E-10, r=1.E+12, hermitian=False):
    """Construct truncated skeleton decomposition A = U V for the given matrix.

    Args:
        A (np.ndarray): matrix of the shape [m, n].
        e (float): desired approximation accuracy (> 0).
        r (int, float): maximum rank for the SVD decomposition (> 0).
        hermitian (flag): if True, then "hermitian" SVD will be used.

    Returns:
        [np.ndarray, np.ndarray]: factor matrix U of the shape [m, q] and factor
        matrix V of the shape [q, n], where "q" is selected rank (q <= r).

    """
    U, s, V = np.linalg.svd(A, full_matrices=False, hermitian=hermitian)
    r = max(1, min(int(r), sum(s > e)))
    S = np.diag(np.sqrt(s[:r]))
    return U[:, :r] @ S, S @ V[:r, :]


def matrix_svd(A, e=1.E-10, r=1.E+12, e0=1.E-14):
    """Construct truncated SVD decomposition A = U V for the given matrix.

    Args:
        A (np.ndarray): matrix of the shape [m, n].
        e (float): desired approximation accuracy (> 0).
        r (int, float): maximum rank for the SVD decomposition (> 0).
        e0 (float): minimum norm of the A.T A matrix (> 0). If norm less than
            this value, then rank-1 zero factors will be returned.

    Returns:
        [np.ndarray, np.ndarray]: factor matrix U of the shape [m, q] and factor
        matrix V of the shape [q, n], where "q" is selected rank (q <= r).

    """
    m, n = A.shape
    C = A @ A.T if m <= n else A.T @ A

    if np.linalg.norm(C) < e0:
        return np.zeros([m, 1]), np.zeros([1, n])

    w, U = np.linalg.eigh(C)

    w[w < 0] = 0.
    w = np.sqrt(w)

    idx = np.argsort(w)[::-1]
    w = w[idx]
    U = U[:, idx]

    s = w**2
    where = np.where(np.cumsum(s[::-1]) <= e**2)[0]
    dlen = 0 if len(where) == 0 else int(1 + where[-1])
    rank = max(1, min(int(r), len(s) - dlen))
    w = w[:rank]
    U = U[:, :rank]

    V = ((1. / w)[:, np.newaxis] * U.T) @ A if m <= n else U.T
    U = U * w if m <= n else A @ U

    return U, V


def svd(Y_full, e=1E-10, r=1.E+12):
    """Construct TT-tensor from the given full tensor using TT-SVD algorithm.

    Args:
        Y_full (np.ndarray): tensor (multidimensional array) in the full format.
        e (float): desired approximation accuracy (> 0).
        r (int, float): maximum rank of the constructed TT-tensor (> 0).

    Returns:
        list: TT-tensor, which represents an approximation with a given
        accuracy (e) and a TT-rank constraint (r) for the given full tensor.

    """
    n = Y_full.shape
    Z = Y_full.copy()
    Y = []
    q = 1
    for k in n[:-1]:
        Z = Z.reshape(q * k, -1)
        G, Z = matrix_skeleton(Z, e, r)
        G = G.reshape(q, k, -1)
        q = G.shape[-1]
        Y.append(G)
    Y.append(Z.reshape(q, n[-1], 1))
    return Y


def svd_incomplete(I, Y, idx, idx_many, e=1.E-10, r=1.E+12):
    """Construct TT-tensor from the given specially selected samples.

    Args:
        I (np.ndarray): multi-indices for the tensor in the form of array
            of the shape [samples, d].
        Y (np.ndarray): values of the tensor for multi-indices I in the form of
            array of the shape [samples].
        idx (np.ndarray): starting poisitions in generated samples for the
            corresponding dimensions in the form of array of the shape [d+1].
        idx_many (np.ndarray): numbers of points for the right unfoldings in
            generated samples in the form of array of the shape [d].
        e (float): desired approximation accuracy (> 0).
        r (int, float): maximum rank of the constructed TT-tensor (> 0).

    Returns:
        list: TT-tensor, which represents an approximation with a given
        accuracy (e) and a TT-rank constraint (r) for the full tensor.

    Note:
        The samples I and opts idx and idx_many should be generated by the
        function "sample_tt".

    """
    shapes = np.max(I, axis=0) + 1
    d = len(shapes)

    Y_curr = Y[idx[0]:idx[1]]
    Y_curr = Y_curr.reshape(shapes[0], -1, order='C')
    Y_curr, _ = matrix_skeleton(Y_curr, e, r)
    Y_res = [Y_curr[None, ...]]

    for mode in range(1, d):
        # The mode-th TT-core will have the shape r0 x n x r1:
        r0 = Y_res[-1].shape[-1]
        r1 = r if mode < d-1 else 1
        n = shapes[mode]

        I_curr = I[idx[mode]:idx[mode+1], :]
        M = np.array([get(Y_res[:mode], i, to_item=False)
            for i in I_curr[::idx_many[mode], :mode]])

        Y_curr = Y[idx[mode]:idx[mode+1]].reshape(-1, idx_many[mode], order='C')
        if Y_curr.shape[1] > r1:
            Y_curr, _ = matrix_skeleton(Y_curr, e, r1)
        r1 = Y_curr.shape[1]

        G = np.empty([r0, n, r1])
        step = Y_curr.shape[0] // n
        for i in range(n):
            A = M[i*step:(i+1)*step]
            b = Y_curr[i*step:(i+1)*step]
            G[:, i, :] = np.linalg.lstsq(A, b, rcond=-1)[0]
        Y_res.append(G)

    return Y_res



"""Package teneva, module core.grid: functions to construct the grid.

This module contains a set of functions for creating and transforming
multidimensional grids for discretizing functions on uniform and Chebyshev
grids, as well as a number of methods for tensor sampling.

"""
import itertools
import numpy as np


def cache_to_data(cache={}):
    I = np.array([str_to_ind(s) for s in cache.keys()], dtype=int)
    Y = np.array([y for y in cache.values()])
    return I, Y


def grid_flat(n):
    """Compute the multi-indices for the full multidimensional grid.

    Args:
        n (list, np.ndarray): number of grid points for each dimension (list or
            np.ndarray of length "d", where "d" is a number of dimensions).

    Returns:
        np.ndarray: multi-indices for the full (flatten) grid (it is 2D array of
        the shape [d, n^d]).

    """
    d = len(n)
    I = [np.arange(k).reshape(1, -1) for k in n]
    I = np.meshgrid(*I, indexing='ij')
    I = np.array(I, dtype=int).reshape((d, -1), order='F').T
    return I


def grid_prep_opt(opt, d=None, kind=float, reps=None):
    """Helper function that prepare grid parameter.

    Args:
        opt (int, float, list, np.ndarray): grid parameter for each dimension.
        d (int): the dimension of the grid. If should be set if "opt" is scalar.
        kind (class 'int', class 'float'): data type for option (int or float).
        reps (int): optional number of repetitions for option.

    Returns:
        np.ndarray: prepared option. It is 1D np.ndarray of length "d" if reps
        is None, or np.ndarray of the shape [reps, d] (values repeated along
        the 1th axis) otherwise.

    """
    if opt is None:
        return None

    if isinstance(opt, (int, float)):
        if d is None or d <= 0:
            raise ValueError('Invalid grid option')
        opt = np.ones(d, dtype=kind) * kind(opt)
    opt = np.asanyarray(opt, dtype=kind)

    if reps is not None:
        opt = np.repeat(opt.reshape((1, -1)), reps, axis=0)

    return opt


def grid_prep_opts(a=None, b=None, n=None, d=None, reps=None):
    """Helper function that prepare grid parameters a, b, n.

    Args:
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length "d" or float).
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length "d" or float).
        n (int, float, list, np.ndarray): grid size for each dimension (list or
            np.ndarray of length "d" or int/float).
        d (int): the dimension of the grid. If all "a", "b", "n" are numbers,
            then it should be set, otherwise it is optional ("d" will be
            recoverd fromv the given "a" / "b" / "n").
        reps (int): optional number of repetitions for a, b, n.

    Returns:
        [np.ndarray, np.ndarray, np.ndarray]: grid lower bounds "a", grid upper
        bounds "b" and grid size "n" for each dimension. All opts will be 1D
        arrays of length "d" if reps is None, or np.ndarray of the shape
        [reps, d] (values repeated along the 1th axis) otherwise.

    Note:
        In case of a mismatch in the size of the arrays "a" / "b" / "n" or the
        impossibility of restoring the dimension "d", the function will
        generate the error (ValueError).

    """
    for item in [a, b, n]:
        if isinstance(item, (list, np.ndarray)):
            if d is None:
                d = len(item)
            elif d != len(item):
                raise ValueError('Invalid grid option')

    a = grid_prep_opt(a, d, float, reps)
    b = grid_prep_opt(b, d, float, reps)
    n = grid_prep_opt(n, d, int, reps)

    return a, b, n


def ind_to_poi(I, a, b, n, kind='uni'):
    """Transform multi-indices (samples) into points of the spatial grid.

    Args:
        I (list, np.ndarray): multi-indices for the tensor in the form of array
            of the shape [samples, d], where "samples" is the number of samples
            and "d" is the dimension of the tensor. For the case of only one
            sample, it may be 1D array or list of length "d".
        a (float, list, np.ndarray): grid lower bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the lower
            bounds for each dimension will be the same.
        b (float, list, np.ndarray): grid upper bounds for each dimension (list
            or np.ndarray of length "d"). It may be also float, then the upper
            bounds for each dimension will be the same.
        n (int, float, list, np.ndarray): tensor size for each dimension (list
            or np.ndarray of length "d"). It may be also int/float, then the
            size for each dimension will be the same.
        kind (str): the grid type, it may be "uni" (uniform grid) and "cheb"
            (Chebyshev grid). In case of the uniform grid, index "0" relates to
            the spatial point "a" and index "n-1" relates to the spatial point
            "b". In case of the Chebyshev grid, index "0" relates to the
            spatial point "b" and index "n-1" relates to the spatial point "a".

    Returns:
        np.ndarray: points of the grid in the form of array of the shape
        [samples, d]. If input "I" is 1D list or np.ndarray (the case of only
        one sample), then function will also return 1D np.ndarray of length "d".

    Note:
        This function may be used for function approximation by low rank tensor
        methods (requested tensor indices should be transformed into spatial
        grid points and then used as an argument for the target function).

    """
    I = np.asanyarray(I, dtype=int)
    d = I.shape[-1]
    m = I.shape[0] if len(I.shape) > 1 else None
    a, b, n = grid_prep_opts(a, b, n, d, m)

    if kind == 'uni':
        X = I / (n - 1) * (b - a) + a
    elif kind == 'cheb':
        X = np.cos(np.pi * I / (n - 1)) * (b - a) / 2 + (b + a) / 2
    else:
        raise ValueError(f'Unknown grid type "{kind}"')

    return X


def ind_to_str(i):
    """Transforms array of int like [1, 2, 3] into string like '1-2-3'.

    Simple function that may be used for the cache of the TT-CROSS.

    Args:
        i (list, np.ndarray): multi-index in the form of array of the shape [d]
            or list of the length "d".

    Returns:
        str: multi-index converted to string, where indexes are separated by
        hyphens.

    """
    return '-'.join([str(int(v)) for v in i])


def sample_lhs(n, m):
    """Generate LHS samples (multi-indices) for the tensor of the given shape.

    Args:
        n (list, np.ndarray): tensor size for each dimension (list or
            np.ndarray of int/float of the length "d", where "d" is the
            dimension of the tensor).
        m (int, float): number of samples.

    Returns:
        np.ndarray: generated multi-indices for the tensor in the form of array
        of the shape [m, d], where "d" is the dimension of the tensor.

    """
    n = np.asanyarray(n, dtype=int)
    m = int(m)
    d = n.shape[0]

    I = np.empty((m, d), dtype=int)
    for i, k in enumerate(n):
        I1 = np.repeat(np.arange(k), m // k)
        I2 = np.random.choice(k, m-len(I1), replace=False)
        I[:, i] = np.concatenate([I1, I2])
        np.random.shuffle(I[:, i])

    return I


def sample_tt(n, r=4):
    """Generate special samples for the tensor of the shape n.

    Generate special samples (multi-indices) for the tensor, which are the best
    (in many cases) for the subsequent construction of the TT-tensor.

    Args:
        n (list, np.ndarray): tensor size for each dimension (list or
            np.ndarray of int/float of the length "d").
        r (int): expected TT-rank of the tensor. The number of generated
            samples will be selected according to this value.

    Returns:
        [np.ndarray, np.ndarray, np.ndarray]: generated multi-indices for the
        tensor in the form of array of the shape [samples, d], starting
        poisitions in generated samples for the corresponding dimensions in the
        form of array of the shape [d+1] and numbers of points for the right
        unfoldings in generated samples in the form of array of the shape [d].

    Note:
        The resulting number of samples will be chosen adaptively based on the
        specified expected TT-rank (r).

    """
    def one_mode(sh1, sh2, rng):
        res = []
        if len(sh2) == 0:
            lhs_1 = sample_lhs(sh1, r)
            for n in range(rng):
                for i in lhs_1:
                    res.append(np.concatenate([i, [n]]))
            len_1, len_2 = len(lhs_1), 1
        elif len(sh1) == 0:
            lhs_2 = sample_lhs(sh2, r)
            for n in range(rng):
                for j in lhs_2:
                    res.append(np.concatenate([[n], j]))
            len_1, len_2 = 1, len(lhs_2)
        else:
            lhs_1 = sample_lhs(sh1, r)
            lhs_2 = sample_lhs(sh2, r)
            for n in range(rng):
                for i, j in itertools.product(lhs_1, lhs_2):
                    res.append(np.concatenate([i, [n], j]))
            len_1, len_2 = len(lhs_1), len(lhs_2)
        return res, len_1, len_2

    I, idx, idx_many = [], [0], []
    for i in range(len(n)):
        pnts, len_1, len_2 = one_mode(n[:i], n[i+1:], n[i])
        I.append(pnts)
        idx.append(idx[-1] + len(pnts))
        idx_many.append(len_2)

    return np.vstack(I), np.array(idx), np.array(idx_many)


def str_to_ind(s):
    """Transforms string like '1-2-3' into array of int like [1, 2, 3].

    Simple function that transforms string like `1-2-3` into array of int like
    `[1, 2, 3]` (it is used for the cache of the TT-cross).

    Args:
        s (str): d-dimensional multi-index in the form of the string, where
            indices are separated by hyphens.

    Returns:
        np.ndarray: multi-index in the form of the 1D array of size "d".

    """
    return np.array([int(v) for v in s.split('-')], dtype=int)



"""Package teneva, module func.demo.func_demo_brown: function.

This module contains class that implements analytical Brown function
for demo and tests.

"""
import numpy as np


from ..func import Func


class FuncDemoBrown(Func):
    def __init__(self, d):
        """Brown function for demo and tests.

        Args:
            d (int): number of dimensions.

        Note:
            See Momin Jamil, Xin-She Yang. "A literature survey of benchmark
            functions for global optimization problems". Journal of
            Mathematical Modelling and Numerical Optimisation 2013; 4:150-194
            for details.

        """
        super().__init__(d, name='Brown')

        self.set_lim(-1., +4.)
        self.set_min([0.]*self.d, 0.)

    def _calc(self, x):
        y = (x[:-1]**2)**(x[1:]**2+1) + (x[1:]**2)**(x[:-1]**2+1)
        return np.sum(y)

    def _comp(self, X):
        Y = (X[:, :-1]**2)**(X[:, 1:]**2+1) + (X[:, 1:]**2)**(X[:, :-1]**2+1)
        return np.sum(Y, axis=1)



"""Package teneva, module core.maxvol: maxvol-like algorithms.

This module contains the implementation of the maxvol algorithm (function
maxvol) and rect_maxvol algorithm (function maxvol_rect) for matrices. The
corresponding functions find in a given matrix square and rectangular
maximal-volume submatrix, respectively (for the case of square submatrix, it
has approximately the maximum value of the modulus of the determinant).

"""
import numpy as np
from scipy.linalg import lu
from scipy.linalg import solve_triangular


def maxvol(A, e=1.05, k=100):
    """Compute the maximal-volume submatrix for given tall matrix.

    Args:
        A (np.ndarray): tall matrix of the shape [n, r] (n > r).
        e (float): accuracy parameter (should be >= 1). If the parameter is
            equal to 1, then the maximum number of iterations will be performed
            until true convergence is achieved. If the value is greater than
            one, the algorithm will complete its work faster, but the accuracy
            will be slightly lower (in most cases, the optimal value is within
            the range of 1.01 - 1.1).
        k (int): maximum number of iterations (should be >= 1).

    Returns:
        [np.ndarray, np.ndarray]: the row numbers "I" containing the
        maximal-volume submatrix in the form of 1D array of length "r" and
        coefficient matrix "B" in the form of 2D array of shape [n, r], such
        that A = B A[I, :] and A (A[I, :])^{-1} = B.

    Note:
        The description of the basic implementation of this algorithm is
        presented in the work: Goreinov S., Oseledets, I., Savostyanov, D.,
        Tyrtyshnikov, E., Zamarashkin, N. "How to find a good submatrix".
        Matrix Methods: Theory, Algorithms And Applications: Dedicated to the Memory of Gene Golub (2010): 247-256.

    """
    n, r = A.shape

    if n <= r:
        raise ValueError('Input matrix should be "tall"')

    P, L, U = lu(A, check_finite=False)
    I = P[:, :r].argmax(axis=0)
    Q = solve_triangular(U, A.T, trans=1, check_finite=False)
    B = solve_triangular(L[:r, :], Q, trans=1, check_finite=False,
        unit_diagonal=True, lower=True).T

    for _ in range(k):
        i, j = np.divmod(np.abs(B).argmax(), r)
        if np.abs(B[i, j]) <= e:
            break

        I[j] = i

        bj = B[:, j]
        bi = B[i, :].copy()
        bi[j] -= 1.

        B -= np.outer(bj, bi / B[i, j])

    return I, B


def maxvol_rect(A, e, dr_min=0, dr_max=None, e0=1.05, k0=10):
    """Compute the maximal-volume rectangular submatrix for given tall matrix.

    Within the framework of this function, the original maxvol algorithm is
    first called (see function "maxvol"). Then additional rows of the matrix
    are greedily added until the maximum allowed number is reached or until
    convergence.

    Args:
        A (np.ndarray): tall matrix of the shape [n, r] (n > r).
        e (float): accuracy parameter.
        dr_min (int): minimum number of added rows (should be >= 0 and <= n-r).
        dr_max (int): maximum number of added rows (should be >= 0). If the
            value is not specified, then the number of added rows will be
            determined by the precision parameter "e", while the resulting
            submatrix can even has the same size as the original matrix "A".
            If r + dr_max is greater than n, then dr_max will be set such that
            r + dr_max = n.
        e0 (float): accuracy parameter for the original maxvol algorithm
            (should be >= 1). See function "maxvol" for details.
        k0 (int): maximum number of iterations for the original maxvol algorithm
            (should be >= 1). See function "maxvol" for details.

    Returns:
        [np.ndarray, np.ndarray]: the row numbers "I" containing the rectangular
        maximal-volume submatrix in the form of 1D array of length r + dr,
        where "dr" is a number of additional selected rows (dr >= dr_min and
        dr <= dr_max) and coefficient matrix "B" in the form of 2D array of
        the shape [n, r+dr], such that A = B A[I, :].

    Note:
        The description of the basic implementation of this algorithm is
        presented in the work: Mikhalev A, Oseledets I., "Rectangular
        maximum-volume submatrices and their applications." Linear Algebra and
        its Applications 538 (2018): 187-211.

    """
    n, r = A.shape
    r_min = r + dr_min
    r_max = r + dr_max if dr_max is not None else n
    r_max = min(r_max, n)

    if r_min < r or r_min > r_max or r_max > n:
        raise ValueError('Invalid minimum/maximum number of added rows')

    I0, B = maxvol(A, e0, k0)

    I = np.hstack([I0, np.zeros(r_max-r, dtype=I0.dtype)])
    S = np.ones(n, dtype=int)
    S[I0] = 0
    F = S * np.linalg.norm(B, axis=1)**2

    for k in range(r, r_max):
        i = np.argmax(F)

        if k >= r_min and F[i] <= e*e:
            break

        I[k] = i
        S[i] = 0

        v = B.dot(B[i])
        l = 1. / (1 + v[i])
        B = np.hstack([B - l * np.outer(v, B[i]), l * v.reshape(-1, 1)])
        F = S * (F - l * v * v)

    I = I[:B.shape[1]]
    B[I] = np.eye(B.shape[1], dtype=B.dtype)

    return I, B



"""Package teneva, module func.func: class that represents the function."""
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import matplotlib.pyplot as plt
from matplotlib.ticker import LinearLocator, FormatStrFormatter
import numpy as np
import pickle
from time import perf_counter as tpc


from teneva import als
from teneva import anova
from teneva import cache_to_data
from teneva import cheb_int
from teneva import cheb_get
from teneva import cheb_gets
from teneva import cross
from teneva import erank
from teneva import getter
from teneva import ind_to_poi
from teneva import rand
from teneva import sample_lhs
from teneva import size


class Func:
    def __init__(self, d, f_calc=None, f_comp=None, name='Demo'):
        """Multivariable function with approximation methods and helpers.

        Class represents the multivariable function (black box) with various
        methods (including TT-ALS, TT-ANOVA and TT-CROSS) for constructing its
        low-rank tensor approximation in the TT-format and utilities for
        building datasets (train, validation, test), check the accuracy, etc.

        Args:
            d (int): number of dimensions of the target function's input.
            f_calc (function): python function that returns the value (float)
                of the target function for the given multidimensional point
                (1D np.ndarray of the length "d"). If it is not set, then the
                corresponding function code is expected to be written in the
                class method "_calc". This function is used to build the
                train, validation and test datasets. If "f_calc" argument and
                "_calc" method are not set, then the "f_comp" / "_comp" method
                will be automatically used for the function evaluation for one
                point.
            f_comp (function): python function that returns the values (1D
                np.ndarray of the length "samples") of the target function for
                the given multidimensional points (2D np.ndarray of the shape
                [samples, d]). If it is not set, then the corresponding
                function code is expected to be written in the class method
                "_comp". This function is used to build the train, validation
                and test datasets. If "f_comp" argument and "_comp" method are
                not set, then the "f_calc" / "_calc" method will be
                automatically used for the function evaluation for the batch of
                points.
            name (str): optional display name for the function.

        Note:
            If "f_calc" / "_calc" / "f_comp" / "_comp" methods are not set,
            then train, validation and test datasets should be set manually by
            "set_trn_ind" / "set_trn_poi", "set_vld_ind" / "set_vld_poi" and
            "set_tst_ind" / "set_tst_poi" methods.

            Class instance method "get_f_poi" should be used for the target
            function evaluation (for one point or for the batch of points). To
            evaluate the target function for the grid multi-index (one or
            batch), use method "get_f_ind" (this function is very usefull for
            tensor approximation methods).

            To evaluate the TT-approximation of the target function for one
            point or for the batch of points, use the "get_poi" method of just
            call the class instance. To evaluate the TT-approximation of the
            target function on the grid multi-index, use "get_ind" method or
            "getitem" notation.

        """
        self.d = d
        self.f_calc = f_calc
        self.f_comp = f_comp
        self.name = name

        self.I_trn_ind = None
        self.I_trn_poi = None
        self.I_tst_ind = None
        self.I_tst_poi = None
        self.I_vld_ind = None
        self.I_vld_poi = None
        self.X_trn_ind = None
        self.X_trn_poi = None
        self.X_tst_ind = None
        self.X_tst_poi = None
        self.X_vld_ind = None
        self.X_vld_poi = None
        self.Y_trn_ind = None
        self.Y_trn_poi = None
        self.Y_tst_ind = None
        self.Y_tst_poi = None
        self.Y_vld_ind = None
        self.Y_vld_poi = None

        self.m_trn_ind = 0
        self.m_trn_poi = 0
        self.m_tst_ind = 0
        self.m_tst_poi = 0
        self.m_vld_ind = 0
        self.m_vld_poi = 0

        self.t_trn_ind_build = 0.
        self.t_trn_poi_build = 0.
        self.t_tst_ind_build = 0.
        self.t_tst_poi_build = 0.
        self.t_vld_ind_build = 0.
        self.t_vld_poi_build = 0.

        self.set_grid()
        self.set_lim()
        self.set_min()
        self.set_noise()
        self.clear()

    def __call__(self, X):
        return self.get_poi(X)

    def __getitem__(self, I):
        return self.get_ind(I)

    def als(self, nswp=30):
        """Build approximation, using TT-ALS.

        See "teneva.core.als" for more details. Initial approximation should
        be prepared before the call of this function.

        """
        self.method = self.method + '-ALS' if self.method else 'ALS'

        if not self.m_trn_ind:
            raise ValueError('Train data is not ready')
        if self.Y is None:
            raise ValueError('Initial approximation is not ready')

        t = tpc()
        Y = als(self.I_trn_ind, self.Y_trn_ind, self.Y, nswp)
        self.t += tpc() - t

        if not self.m:
            self.m = self.m_trn_ind
        self.nswp = nswp

        self.prep(Y)

    def anova(self, r=2, order=1):
        """Build approximation, using TT-ANOVA.

        See "teneva.core.anova" for more details.

        """
        self.method = self.method + '-ANO' if self.method else 'ANO'

        if not self.m_trn_ind:
            raise ValueError('Train data is not ready')

        t = tpc()
        Y = anova(self.I_trn_ind, self.Y_trn_ind, r, order)
        self.t += tpc() - t

        if not self.m:
            self.m = self.m_trn_ind

        self.prep(Y)

    def build_trn_ind(self, m):
        """Generate train dataset from random indices.

        Args:
            m (int or float): number of points to generate.

        Note:
            Class instance variables I_trn_ind (grid indices), X_trn_ind
            (spatial points related to grid indices) and Y_trn_ind (function
            values) will be built. Indices I_trn_ind will be generated from the
            LHS distribution. If the noise parameters are not None, then the
            corresponding random noise will be added for the data.

        """
        t = tpc()
        a, b, n, f = self.a, self.b, self.n, sample_lhs
        self.m_trn_ind = int(m)
        self.I_trn_ind = f(n, self.m_trn_ind)
        self.X_trn_ind = ind_to_poi(self.I_trn_ind, a, b, n, self.kind)
        self.Y_trn_ind = self.get_f_poi_spec(self.X_trn_ind)
        self.t_trn_ind_build = tpc() - t

    def build_trn_poi(self, m):
        """Generate train dataset from spatial points.

        Args:
            m (int or float): number of points to generate.

        Note:
            Class instance variables X_trn_poi (spatial points) and Y_trn_poi
            (function values) will be built. If the noise parameters are not
            None, then the corresponding random noise will be added for the
            data.

        """
        t = tpc()
        a, b, f = self.a, self.b, np.random.uniform
        self.m_trn_poi = int(m)
        self.I_trn_poi = None
        self.X_trn_poi = np.vstack([f(a, b) for _ in range(self.m_trn_poi)])
        self.Y_trn_poi = self.get_f_poi_spec(self.X_trn_poi)
        self.t_trn_poi_build = tpc() - t

    def build_tst_ind(self, m):
        """Generate test dataset from random tensor indices.

        Args:
            m (int or float): number of points to generate.

        Note:
            Class instance variables I_tst_ind (grid indices), X_tst_ind
            (spatial points related to grid indices) and Y_tst_ind (function
            values) will be built. Indices I_trn_ind will be generated from
            uniform random choice of indices.

        """
        t = tpc()
        a, b, n, f = self.a, self.b, self.n, np.random.choice
        self.m_tst_ind = int(m)
        self.I_tst_ind = np.vstack([f(k, self.m_tst_ind) for k in n]).T
        self.X_tst_ind = ind_to_poi(self.I_tst_ind, a, b, n, self.kind)
        self.Y_tst_ind = self.get_f_poi(self.X_tst_ind)
        self.t_tst_ind_build = tpc() - t

    def build_tst_poi(self, m):
        """Generate test dataset from spatial points.

        Args:
            m (int or float): number of points to generate.

        Note:
            Class instance variables X_tst_poi (spatial points) and Y_tst_poi
            (function values) will be built.

        """
        t = tpc()
        a, b, f = self.a, self.b, np.random.uniform
        self.m_tst_poi = int(m)
        self.I_tst_poi = None
        self.X_tst_poi = np.vstack([f(a, b) for _ in range(self.m_tst_poi)])
        self.Y_tst_poi = self.get_f_poi(self.X_tst_poi)
        self.t_tst_poi_build = tpc() - t

    def build_vld_ind(self, m):
        """Generate validation dataset from random tensor indices.

        Args:
            m (int or float): number of points to generate.

        Note:
            Class instance variables I_vld_ind (grid indices), X_vld_ind
            (spatial points related to grid indices) and Y_vld_ind (function
            values) will be built. Indices I_vld_ind will be generated from
            uniform random choice of indices. If the noise parameters are not
            None, then the corresponding random noise will be added for the
            data.

        """
        t = tpc()
        a, b, n, f = self.a, self.b, self.n, np.random.choice
        self.m_vld_ind = int(m)
        self.I_vld_ind = np.vstack([f(k, self.m_vld_ind) for k in n]).T
        self.X_vld_ind = ind_to_poi(self.I_vld_ind, a, b, n, self.kind)
        self.Y_vld_ind = self.get_f_poi_spec(self.X_vld_ind)
        self.t_vld_ind_build = tpc() - t

    def build_vld_poi(self, m):
        """Generate validation dataset from spatial points.

        Args:
            m (int or float): number of points to generate.

        Note:
            Class instance variables X_vld_poi (spatial points) and Y_vld_poi
            (function values) will be built. If the noise parameters are not
            None, then the corresponding random noise will be added for the
            data.

        """
        t = tpc()
        a, b, f = self.a, self.b, np.random.uniform
        self.m_vld_poi = int(m)
        self.I_vld_poi = None
        self.X_vld_poi = np.vstack([f(a, b) for _ in range(self.m_vld_poi)])
        self.Y_vld_poi = self.get_f_poi_spec(self.X_vld_poi)
        self.t_vld_poi_build = tpc() - t

    def check(self):
        """Compute the error of the TT-approximation on all datasets."""
        if self.m_trn_ind:
            self.check_trn_ind()
        if self.m_trn_poi:
            self.check_trn_poi()
        if self.m_tst_ind:
            self.check_tst_ind()
        if self.m_tst_poi:
            self.check_tst_poi()
        if self.m_vld_ind:
            self.check_vld_ind()
        if self.m_vld_poi:
            self.check_vld_poi()

    def check_trn_ind(self):
        """Calculate the TT-approximation error for train indices."""
        self.e_trn_ind, self.t_trn_ind_check = _data_check(
            self.I_trn_ind, self.Y_trn_ind, self.get_ind)
        return self.e_trn_ind

    def check_trn_poi(self):
        """Calculate the TT-approximation error for train points."""
        if self.kind != 'cheb':
            raise ValueError('Can check only "cheb" spatial points')

        self.e_trn_poi, self.t_trn_poi_check = _data_check(
            self.X_trn_poi, self.Y_trn_poi, self.get_poi)
        return self.e_trn_poi

    def check_tst_ind(self):
        """Calculate the TT-approximation error for test indices."""
        self.e_tst_ind, self.t_tst_ind_check = _data_check(
            self.I_tst_ind, self.Y_tst_ind, self.get_ind)
        return self.e_tst_ind

    def check_tst_poi(self):
        """Calculate the TT-approximation error for test points."""
        if self.kind != 'cheb':
            raise ValueError('Can check only "cheb" spatial points')

        self.e_tst_poi, self.t_tst_poi_check = _data_check(
            self.X_tst_poi, self.Y_tst_poi, self.get_poi)
        return self.e_tst_poi

    def check_vld_ind(self):
        """Calculate the TT-approximation error for validation indices."""
        self.e_vld_ind, self.t_vld_ind_check = _data_check(
            self.I_vld_ind, self.Y_vld_ind, self.get_ind)
        return self.e_vld_ind

    def check_vld_poi(self):
        """Calculate the TT-approximation error for validation points."""
        if self.kind != 'cheb':
            raise ValueError('Can check only "cheb" spatial points')

        self.e_vld_poi, self.t_vld_poi_check = _data_check(
            self.X_vld_poi, self.Y_vld_poi, self.get_poi)
        return self.e_vld_poi

    def clear(self):
        """Remove all results of the previous approximations.

        Note:
            Note that this function does not remove the datasets. So, for
            example, TT-CROSS cache (train data) may be used in the following
            approximation.

        """
        self.Y = None             # TT-tensor of function values on the grid
        self.r = 0                # Effective TT-rank for Y

        self._get = None          # Getter for the TT-tensor
        self.A = None             # TT-tensor of interpolation coefficients

        self.t = 0.               # Total time of approximation
        self.e = -1.              # Expected error of approximation
        self.m = 0                # Total number of requests to target function
        self.m_cache = 0          # Total number of requests to cache
        self.method = ''          # Name of the approximation method
        self.nswp = 0             # Number of (CROSS or ALS) sweeps
        self.stop = 'm'           # Stop condition (CROSS) for approximation

        self.e_trn_ind = -1.      # Relative errors on the datasets
        self.e_trn_poi = -1.
        self.e_tst_ind = -1.
        self.e_tst_poi = -1.
        self.e_vld_ind = -1.
        self.e_vld_poi = -1.

        self.t_trn_ind_check = 0. # Time of error check for datasets
        self.t_trn_poi_check = 0.
        self.t_tst_ind_check = 0.
        self.t_tst_poi_check = 0.
        self.t_vld_ind_check = 0.
        self.t_vld_poi_check = 0.

    def cross(self, m=None, e=1.E-8, nswp=100, dr_min=1, dr_max=2, cache=True):
        """Build approximation, using TT-cross (TT-CAM).

        See "teneva.core.cross" for more details. Initial approximation should
        be prepared before the call of this function.

        """
        self.method = self.method + '-CRO' if self.method else 'CRO'

        if self.Y is None:
            raise ValueError('Initial approximation is not ready')

        t = tpc()
        info = {}
        cache = {} if cache else None
        f = self.get_f_ind_spec
        Y = cross(f, self.Y, e, m, nswp, dr_min, dr_max, info, cache)
        self.t += tpc() - t

        self.m += info['m']
        self.m_cache += info['m_cache']
        self.e = info['e']
        self.nswp = info['nswp']
        self.stop = info['stop']

        if cache is not None and len(cache) > 0:
            I_, Y_ = cache_to_data(cache)
            X_ = ind_to_poi(I_, self.a, self.b, self.n, self.kind)
            self.set_trn_ind(I_, X_, Y_)

        self.prep(Y)

    def get_f_ind(self, I):
        """Calculate the target function in the given grid multi-index.

        Args:
            I (list, np.ndarray): grid index in the form of array of the shape
                [d], where "d" is the dimension of the input or batch of grid
                indices in the form of array of the shape [samples, d], where
                "samples" is the number of samples.

        Returns:
            float or np.ndarray: the function value in the point related to
            the given grid index or the values of the function in the points
            related to given grid indices in the form of array of the shape
            [samples].

        Note:
            Grid parameters may be set using "set_grid" function.

        """
        X = ind_to_poi(I, self.a, self.b, self.n, self.kind)
        return self.get_f_poi(X)

    def get_f_ind_spec(self, I):
        """Compute the target function values for indices with special options.

        Args:
            I (list, np.ndarray): grid multi-indices in the form of array of
                the shape [samples, d], where "samples" is the number of
                samples and "d" is the dimension of the input.

        Returns:
            float or np.ndarray: the values of the function in the points
            related to given grid indices in the form of array of the shape
            [samples].

        Note:
            If noise parameters are set (see "set_noise" function), then noise
            will be added to the result of this function.

        """
        y = self.get_f_ind(I)
        y = _noise(y, self.noise_add, self.noise_mul)
        return y

    def get_f_poi(self, X):
        """Calculate the target function in the given point or points.

        Args:
            X (list, np.ndarray): point (function input) in the form of array
                of the shape [d], where "d" is the dimension of the input or
                batch of points (function inputs) in the form of array of the
                shape [samples, d], where "samples" is the number of samples.

        Returns:
            float or np.ndarray: the value of the target function in given
            point or the values in given points in the form of array of the
            shape [samples].

        """
        X = np.asanyarray(X, dtype=float)

        if len(X.shape) == 1:
            return self._calc(X)
        else:
            return self._comp(X)

    def get_f_poi_spec(self, X):
        """Compute the target function values for points with special options.

        Args:
            X (list, np.ndarray): spatial points in the form of array of
                the shape [samples, d], where "samples" is the number of
                samples and "d" is the dimension of the input.

        Returns:
            float or np.ndarray: the values of the function in the given points
            in the form of array of the shape [samples].

        Note:
            If noise parameters are set (see "set_noise" function), then noise
            will be added to the result of this function.

        """
        y = self.get_f_poi(X)
        y = _noise(y, self.noise_add, self.noise_mul)
        return y

    def get_ind(self, I):
        """Calculate the approximation in the given grid multi-index.

        Args:
            I (list, np.ndarray): grid index in the form of array of the shape
                [d], where "d" is the dimension of the input or batch of grid
                indices in the form of array of the shape [samples, d], where
                "samples" is the number of samples.

        Returns:
            float or np.ndarray: the value of the approximation for the target
            function in the point related to given grid index or the values in
            the points related to given grid indices in the form of array of
            the shape [samples].

        Note:
            Grid parameters may be set using "set_grid" function.

        """
        if self._get is None:
            raise ValueError('Approximation is not ready')

        I = np.asanyarray(I, dtype=int)

        if len(I.shape) == 1:
            return self._get(I)
        else:
            return np.array([self._get(i) for i in I])

    def get_poi(self, X):
        """Calculate the approximation of function in the given point or points.

        Args:
            X (list, np.ndarray): point (function input) in the form of array
                of the shape [d], where "d" is the dimension of the input or
                batch of points (function inputs) in the form of array of the
                shape [samples, d], where "samples" is the number of samples.

        Returns:
            float or np.ndarray: the value of the approximation for the target
            function in given point or the values in given points in the form
            of array of the shape [samples].

        """
        if self.Y is None:
            raise ValueError('Approximation is not ready')

        if self.A is None:
            raise ValueError('Interpolation coefficients are not ready')

        X = np.asanyarray(X, dtype=float)

        if len(X.shape) == 1:
            return cheb_get(X.reshape(1, -1), self.A, self.a, self.b)[0]
        else:
            return cheb_get(X, self.A, self.a, self.b)

    def info(self, text_post='', full=False):
        """Present (print) the info about approximation result.

        Note:
            Errors (if available) are presented in the following order: trn
            ind, trn poi, vld ind, vld poi, tst ind, tst poi.

        """
        if full:
            return self.info_full()

        text = ''

        name = self.name or ''
        meth = self.method or ''

        text += name + ' ' * max(0, 15-len(name)) + ' ['
        text += meth + ' ' * max(0, 12-len(meth)) + ' ] > '

        text += f'error: '
        if self.e_trn_ind >= 0:
            text += f'{self.e_trn_ind:-7.1e} / '
        if self.e_trn_poi >= 0:
            text += f'{self.e_trn_poi:-7.1e} / '
        if self.e_vld_ind >= 0:
            text += f'{self.e_vld_ind:-7.1e} / '
        if self.e_vld_poi >= 0:
            text += f'{self.e_vld_poi:-7.1e} / '
        if self.e_tst_ind >= 0:
            text += f'{self.e_tst_ind:-7.1e} / '
        if self.e_tst_poi >= 0:
            text += f'{self.e_tst_poi:-7.1e} / '
        text = text[:-2] + '| '

        text += f'rank: {self.r:-4.1f} | '

        text += f'time: {self.t:-7.3f}'

        if text_post:
            text += ' | ' + text_post

        print(text)

    def info_full(self):
        """Present (print) the full info about approximation result."""
        text = ''
        text += f'=' * 50 + '\n'

        text += f'------------------- | {self.name} function\n'
        if self.method:
            text += f'Method              : {self.method:s}\n'

        text += '\n'

        if self.m > 0:
            text += f'Evals function      : {self.m:-7.1e}\n'
        if self.m_cache > 0:
            text += f'Evals cache         : {self.m_cache:-7.1e}\n'
        if self.r:
            text += f'TT-rank             : {self.r:-7.1f}\n'
        if self.Y is not None:
            text += f'Number of params    : {size(self.Y):-7.1e}\n'

        text += '\n'

        if self.m_trn_ind > 0:
            text += f'Samples trn ind     : {self.m_trn_ind:-7.1e}\n'
        if self.m_trn_poi > 0:
            text += f'Samples trn poi     : {self.m_trn_poi:-7.1e}\n'
        if self.m_vld_ind > 0:
            text += f'Samples vld ind     : {self.m_vld_ind:-7.1e}\n'
        if self.m_vld_poi > 0:
            text += f'Samples vld poi     : {self.m_vld_poi:-7.1e}\n'
        if self.m_tst_ind > 0:
            text += f'Samples tst ind     : {self.m_tst_ind:-7.1e}\n'
        if self.m_tst_poi > 0:
            text += f'Samples tst poi     : {self.m_tst_poi:-7.1e}\n'

        text += '\n'

        if self.e_trn_ind >= 0:
            text += f'Error trn ind       : {self.e_trn_ind:-7.1e}\n'
        if self.e_trn_poi >= 0:
            text += f'Error trn poi       : {self.e_trn_poi:-7.1e}\n'
        if self.e_vld_ind >= 0:
            text += f'Error vld ind       : {self.e_vld_ind:-7.1e}\n'
        if self.e_vld_poi >= 0:
            text += f'Error vld poi       : {self.e_vld_poi:-7.1e}\n'
        if self.e_tst_ind >= 0:
            text += f'Error tst ind       : {self.e_tst_ind:-7.1e}\n'
        if self.e_tst_poi >= 0:
            text += f'Error tst poi       : {self.e_tst_poi:-7.1e}\n'

        text += '\n'

        if self.t > 0:
            text += f'Time approximation  : {self.t:-7.3f}\n'

        if self.t_trn_ind_build > 0:
            text += f'Time trn build ind  : {self.t_trn_ind_build:-7.3f}\n'
        if self.t_trn_ind_check > 0:
            text += f'Time trn check ind  : {self.t_trn_ind_check:-7.3f}\n'
        if self.t_trn_poi_build > 0:
            text += f'Time trn build poi  : {self.t_trn_poi_build:-7.3f}\n'
        if self.t_trn_poi_check > 0:
            text += f'Time trn check poi  : {self.t_trn_poi_check:-7.3f}\n'

        if self.t_vld_ind_build > 0:
            text += f'Time vld build ind  : {self.t_vld_ind_build:-7.3f}\n'
        if self.t_vld_ind_check > 0:
            text += f'Time vld check ind  : {self.t_vld_ind_check:-7.3f}\n'
        if self.t_vld_poi_build > 0:
            text += f'Time vld build poi  : {self.t_vld_poi_build:-7.3f}\n'
        if self.t_vld_poi_check > 0:
            text += f'Time vld check poi  : {self.t_vld_poi_check:-7.3f}\n'

        if self.t_tst_ind_build > 0:
            text += f'Time tst build ind  : {self.t_tst_ind_build:-7.3f}\n'
        if self.t_tst_ind_check > 0:
            text += f'Time tst check ind  : {self.t_tst_ind_check:-7.3f}\n'
        if self.t_tst_poi_build > 0:
            text += f'Time tst build poi  : {self.t_tst_poi_build:-7.3f}\n'
        if self.t_tst_poi_check > 0:
            text += f'Time tst check poi  : {self.t_tst_poi_check:-7.3f}\n'

        text += '\n'

        if self.nswp > 0:
            text += f'Sweeps              : {self.nswp:-7.0f}\n'

        text += f'=' * 50

        print(text)

    def load_trn_ind(self, fpath, m=None):
        """Load the train dataset indices of size m from the npz-file.

        Note:
            If the noise parameters are not None, then the corresponding random
            noise will be added for the loaded data.

        """
        self.set_trn_ind(*_data_load(fpath, m, self.n, self.kind))
        self.Y_trn_ind = _noise(self.Y_trn_ind, self.noise_add, self.noise_mul)

    def load_trn_poi(self, fpath, m=None):
        """Load the train dataset points of size m from the npz-file.

        Note:
            If the noise parameters are not None, then the corresponding random
            noise will be added for the loaded data.

        """
        self.set_trn_poi(*_data_load(fpath, m, self.n, self.kind))
        self.Y_trn_poi = _noise(self.Y_trn_poi, self.noise_add, self.noise_mul)

    def load_tst_ind(self, fpath, m=None):
        """Load the test dataset indices of size m from the npz-file."""
        self.set_tst_ind(*_data_load(fpath, m, self.n, self.kind))

    def load_tst_poi(self, fpath, m=None):
        """Load the test dataset points of size m from the npz-file."""
        self.set_tst_poi(*_data_load(fpath, m, self.n, self.kind))

    def load_vld_ind(self, fpath, m=None):
        """Load the validation dataset indices of size m from the npz-file.

        Note:
            If the noise parameters are not None, then the corresponding random
            noise will be added for the loaded data.

        """
        self.set_vld_ind(*_data_load(fpath, m, self.n, self.kind))
        self.Y_vld_ind = _noise(self.Y_vld_ind, self.noise_add, self.noise_mul)

    def load_vld_poi(self, fpath, m=None):
        """Load the validation dataset points of size m from the npz-file.

        Note:
            If the noise parameters are not None, then the corresponding random
            noise will be added for the loaded data.

        """
        self.set_vld_poi(*_data_load(fpath, m, self.n, self.kind))
        self.Y_vld_poi = _noise(self.Y_vld_poi, self.noise_add, self.noise_mul)

    def plot(self, k=1000):
        """Plot the target function for the 2D case.

        Args:
            k (int): number of points for each dimension.

        """
        if self.d != 2:
            raise ValueError('Plot is supported only for 2D case')

        X1 = np.linspace(self.a[0], self.b[0], k)
        X2 = np.linspace(self.a[1], self.b[1], k)
        X1, X2 = np.meshgrid(X1, X2)
        X = np.hstack([X1.reshape(-1, 1), X2.reshape(-1, 1)])

        Y = self.get_f_poi(X)
        Y = Y.reshape(X1.shape)

        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca(projection='3d')
        ax.set_title(self.name + ' function')
        surf = ax.plot_surface(X1, X2, Y, cmap=cm.coolwarm,
            linewidth=0, antialiased=False)
        ax.zaxis.set_major_locator(LinearLocator(10))
        ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))
        fig.colorbar(surf, shrink=0.3, aspect=10)
        plt.show()

    def prep(self, Y, with_int=True):
        """Prepare getter and interpolation coefficients for approximation.

        Args:
            Y (list): TT-tensor with function values on the grid.
            with_int (bool): if flag is set, then the getter for tensor and
                Chebyshev interpolation coefficients (in the case if Chebyshev
                grid is used) will be prepared.

        """
        if Y is None:
            raise ValueError('Approximation is not ready')

        self.Y = Y
        self.r = erank(self.Y)

        if with_int:
            self._get = getter(self.Y)
            self.A = cheb_int(self.Y) if self.kind == 'cheb' else None
        else:
            self._get = None
            self.A = None

    def rand(self, r=2):
        """Build random approximation in the TT-format.

        Note:
            This function should be called before "als" and "cross" methods
            to prepare the initial approximation for TT-ALS and TT-CROSS
            methods, if other approaches (for example, TT-ANOVA from function
            "anova") are not used to construct the initial approximation.

            Note that interpolation parameters ("_get", "A") will not be
            prepared while call of this function in contrast to the case of
            using other approximation methods ("als", "anova", "cross", etc.).

        """
        self.method = ''

        t = tpc()
        Y = rand(self.n, r)
        self.t += tpc() - t

        self.prep(Y, with_int=False)

    def save_trn_ind(self, fpath):
        """Save train dataset indices to the npz-file."""
        _data_save(fpath, self.I_trn_ind, self.X_trn_ind, self.Y_trn_ind,
            self.t_trn_ind_build, self.n, self.kind)

    def save_trn_poi(self, fpath):
        """Save train dataset points to the npz-file."""
        _data_save(fpath, self.I_trn_poi, self.X_trn_poi, self.Y_trn_poi,
            self.t_trn_poi_build, self.n, self.kind)

    def save_tst_ind(self, fpath):
        """Save test dataset indices to the npz-file."""
        _data_save(fpath, self.I_tst_ind, self.X_tst_ind, self.Y_tst_ind,
            self.t_tst_ind_build, self.n, self.kind)

    def save_tst_poi(self, fpath):
        """Save test dataset points to the npz-file."""
        _data_save(fpath, self.I_tst_poi, self.X_tst_poi, self.Y_tst_poi,
            self.t_tst_poi_build, self.n, self.kind)

    def save_vld_ind(self, fpath):
        """Save validation dataset indices to the npz-file."""
        _data_save(fpath, self.I_vld_ind, self.X_vld_ind, self.Y_vld_ind,
            self.t_vld_ind_build, self.n, self.kind)

    def save_vld_poi(self, fpath):
        """Save validation dataset points to the npz-file."""
        _data_save(fpath, self.I_vld_poi, self.X_vld_poi, self.Y_vld_poi,
            self.t_vld_poi_build, self.n, self.kind)

    def set_grid(self, n=10, kind='cheb'):
        """Set grid options for function discretization.

        Args:
            n (int, float, list, np.ndarray): grid size for each dimension
                (list or np.ndarray of length "d"). It may be also int/float,
                then the size for each dimension will be the same.
            kind (str): the grid kind, it may be "uni" (uniform grid) and "cheb"
                (Chebyshev grid).

        Note:
            If a grid type other than "cheb" is specified, then it will be
            impossible to calculate the value of the approximation result at an
            arbitrary point (method "get_poi") but the calculation in an
            arbitrary tensor multi-index (method "get_ind") will be still
            available.

        """
        if isinstance(n, (int, float)):
            n = np.ones(self.d, dtype=int) * int(n)
        self.n = np.asanyarray(n, dtype=int)
        if len(self.n) != self.d:
            raise ValueError('Invalid length of n')

        if not kind in ['uni', 'cheb']:
            raise ValueError(f'Unknown grid type "{kind}"')
        self.kind = kind

    def set_lim(self, a=-1., b=+1.):
        """Set bounds for the spatial rectangular region.

        Args:
            a (float, list, np.ndarray): lower bounds for each dimension (list
                or np.ndarray of length "d"). It may be also float, then the
                lower bounds for each dimension will be the same.
            b (float, list, np.ndarray): upper bounds for each dimension (list
                or np.ndarray of length "d"). It may be also float, then the
                upper bounds for each dimension will be the same.

        """
        if isinstance(a, (int, float)):
            a = [a] * self.d
        self.a = np.asanyarray(a, dtype=float)

        if isinstance(b, (int, float)):
            b = [b] * self.d
        self.b = np.asanyarray(b, dtype=float)

    def set_min(self, x_min=None, y_min=None):
        """Set the exact global minimum of the function.

        Args:
            x_min (list, np.ndarray): argument of the function that provides its
                minimum value (list or np.ndarray of the length "d").
            y_min (float): minimum value of the function (y_min = f(x_min)).

        """
        if x_min is not None:
            x_min = np.asanyarray(x_min, dtype=float)
        self.x_min = x_min

        self.y_min = y_min

    def set_noise(self, noise_add=None, noise_mul=None):
        """Set the noise parameters.

        If the noise parameters are not None, then the corresponding random
        noise will be added for train and validation data (see functions
        "get_f_ind_spec" and "get_f_poi_spec"; and "build_trn_ind",
        "build_trn_poi", "build_vld_ind", "build_vld_poi"). The TT-CROSS
        method will also use noise data (see function "cross").

        Args:
            noise_add (float): additive noise parameter.
            noise_mul (float): multiplicative noise parameter.

        """
        self.noise_add = noise_add
        self.noise_mul = noise_mul

    def set_trn_ind(self, I=None, X=None, Y=None, t=0.):
        """Set train data indices (indices, points, values and time)."""
        res = _data_prep(I, X, Y)
        self.I_trn_ind, self.X_trn_ind, self.Y_trn_ind, self.m_trn_ind = res
        self.t_trn_ind_build = t

    def set_trn_poi(self, I=None, X=None, Y=None, t=0.):
        """Set train data points (indices, points, values and time)."""
        res = _data_prep(I, X, Y)
        self.I_trn_poi, self.X_trn_poi, self.Y_trn_poi, self.m_trn_poi = res
        self.t_trn_poi_build = t

    def set_tst_ind(self, I=None, X=None, Y=None, t=0.):
        """Set test data indices (indices, points, values and time)."""
        res = _data_prep(I, X, Y)
        self.I_tst_ind, self.X_tst_ind, self.Y_tst_ind, self.m_tst_ind = res
        self.t_tst_ind_build = t

    def set_tst_poi(self, I=None, X=None, Y=None, t=0.):
        """Set test data points (indices, points, values and time)."""
        res = _data_prep(I, X, Y)
        self.I_tst_poi, self.X_tst_poi, self.Y_tst_poi, self.m_tst_poi = res
        self.t_tst_poi_build = t

    def set_vld_ind(self, I=None, X=None, Y=None, t=0.):
        """Set validation data indices (indices, points, values and time)."""
        res = _data_prep(I, X, Y)
        self.I_vld_ind, self.X_vld_ind, self.Y_vld_ind, self.m_vld_ind = res
        self.t_vld_ind_build = t

    def set_vld_poi(self, I=None, X=None, Y=None, t=0.):
        """Set validation data points (indices, points, values and time)."""
        res = _data_prep(I, X, Y)
        self.I_vld_poi, self.X_vld_poi, self.Y_vld_poi, self.m_vld_poi = res
        self.t_vld_poi_build = t

    def _calc(self, x):
        if self.f_calc is None:
            return self._comp(x.reshape(1, -1))[0]
        else:
            return self.f_calc(x)

    def _comp(self, X):
        if self.f_comp is None:
            return np.array([self._calc(x) for x in X])
        else:
            return self.f_comp(X)


def _data_check(X, Y_real, func):
    if X is None:
        return -1., 0.
    t = tpc()
    Y_appr = func(X)
    e = np.linalg.norm(Y_appr - Y_real) / np.linalg.norm(Y_real)
    t = tpc() - t
    return e, t


def _data_load(fpath, m=None, n_ref=None, kind_ref=None):
    data = np.load(fpath, allow_pickle=True)
    I = data.get('I')
    X = data.get('X')
    Y = data.get('Y')
    t = data.get('t').item()
    n = data.get('n')
    kind = data.get('kind').item()

    if n_ref is not None:
        if len(n) != len(n_ref):
            raise ValueError('Invalid dimension for the loaded data')

        for [n_, n_ref_] in zip(n, n_ref):
            if n_ != n_ref_:
                raise ValueError('Invalid grid size for the loaded data')

    if kind_ref is not None:
        if kind != kind_ref:
            raise ValueError('Invalid grid kind for the loaded data')

    if I is not None and len(I.shape) < 2:
        I = None
    if X is not None and len(X.shape) < 2:
        X = None
    if Y is not None and len(Y.shape) < 1:
        Y = None

    if m is not None:
        m = int(m)
        if m > Y.size:
            raise ValueError('Invalid subset of data')

        I = I[:m, :] if I is not None else None
        X = X[:m, :] if X is not None else None
        Y = Y[:m] if Y is not None else None

    return I, X, Y, t


def _data_prep(I=None, X=None, Y=None, perm=None):
    I_new = None
    if I is not None:
        I_new = np.asanyarray(I, dtype=int).copy()
        if perm is not None:
            I_new = I_new[:, perm]

    X_new = None
    if X is not None:
        X_new = np.asanyarray(X, dtype=float).copy()
        if perm is not None:
            X_new = X_new[:, perm]

    Y_new = None
    m_new = 0
    if Y is not None:
        Y_new = np.asanyarray(Y, dtype=float).copy()
        m_new = len(Y_new)

    return I_new, X_new, Y_new, m_new


def _data_save(fpath, I=None, X=None, Y=None, t=None, n=None, kind=None):
    np.savez_compressed(fpath, I=I, X=X, Y=Y, t=t, n=n, kind=kind)


def _noise(y, noise_add=None, noise_mul=None):
    if y is not None and noise_mul is not None:
        y *= noise_mul * np.random.randn(y.size) + 1

    if y is not None and noise_add is not None:
        y += noise_add * np.random.randn(y.size)

    return y



"""Package teneva, module func.demo.func_demo_schaffer: function.

This module contains class that implements analytical Schaffer function
for demo and tests.

"""
import numpy as np


from ..func import Func


class FuncDemoSchaffer(Func):
    def __init__(self, d):
        """Schaffer function for demo and tests.

        Args:
            d (int): number of dimensions.

        Note:
            See Momin Jamil, Xin-She Yang. "A literature survey of benchmark
            functions for global optimization problems". Journal of
            Mathematical Modelling and Numerical Optimisation 2013; 4:150-194
            for details.

        """
        super().__init__(d, name='Schaffer')

        self.set_lim(-100., +100.)
        self.set_min([0.]*self.d, 0.)

    def _calc(self, x):
        z = x[:-1]**2 + x[1:]**2
        y = 0.5 + (np.sin(np.sqrt(z))**2 - 0.5) / (1. + 0.001 * z)**2
        return np.sum(y)

    def _comp(self, X):
        Z = X[:, :-1]**2 + X[:, 1:]**2
        Y = 0.5 + (np.sin(np.sqrt(Z))**2 - 0.5) / (1. + 0.001 * Z)**2
        return np.sum(Y, axis=1)



"""Package teneva, module core.anova: ANOVA decomposition in the TT-format.

This module contains the function "anova" which computes the TT-approximation
for the tensor, using given random samples.

"""
import numpy as np


from .svd import matrix_skeleton
from .tensor import add_many


class ANOVA:
    def __init__(self, I_trn, Y_trn, order=1):
        if not order in [1, 2]:
            raise ValueError('Invalid value for ANOVA order (should be 1 or 2')
        self.order = order

        self.build(I_trn, Y_trn)

    def __call__(self, I):
        I = np.asanyarray(I, dtype=int)
        if len(I.shape) == 1:
            return self.calc(I)
        if len(I.shape) == 2:
            return np.array([self.calc(i) for i in I])
        return None

    def __getitem__(self, I):
        return self(I)

    def build(self, I_trn, Y_trn):
        I_trn = np.asanyarray(I_trn, dtype=int)
        Y_trn = np.asanyarray(Y_trn, dtype=float)
        self.d = I_trn.shape[1]

        self.domain = []
        self.shapes = np.zeros(self.d, dtype=int)
        for k in range(self.d):
            points = np.unique(I_trn[:, k])
            self.domain.append(points)
            self.shapes[k] = len(points)

        self.build_0(I_trn, Y_trn)
        if self.order >= 1:
            self.build_1(I_trn, Y_trn)
        if self.order >= 2:
            self.build_2(I_trn, Y_trn)

    def build_0(self, I_trn, Y_trn):
        self.f0 = np.mean(Y_trn)

    def build_1(self, I_trn, Y_trn):
        self.f1 = []
        self.f1_arr = []
        for k, dm in enumerate(self.domain):
            f1_curr = {}
            f1_curr_arr = []
            for x in dm:
                idx = I_trn[:, k] == x
                value = np.mean(Y_trn[idx]) - self.f0
                f1_curr[x] = value
                f1_curr_arr.append(value)
            self.f1.append(f1_curr)
            self.f1_arr.append(np.array(f1_curr_arr))

    def build_2(self, I_trn, Y_trn):
        self.f2 = []
        self.f2_arr = []
        for k1, dm1 in enumerate(self.domain[:-1]):
            for k2, dm2 in enumerate(self.domain[k1 + 1:], start=k1+1):
                f2_curr = {}
                f2_curr_arr = []
                for x1 in dm1:
                    for x2 in dm2:
                        idx = (I_trn[:, k1] == x1) & (I_trn[:, k2] == x2)
                        if idx.sum() == 0:
                            value = 0.
                        else:
                            df = np.mean(Y_trn[idx]) - self.f0
                            value = df - self.f1[k1][x1] - self.f1[k2][x2]
                        f2_curr[(x1, x2)] = value
                        f2_curr_arr.append(value)
                self.f2.append(f2_curr)
                self.f2_arr.append(np.array(f2_curr_arr))

    def calc(self, i):
        res = self.calc_0()
        if self.order >= 1:
            res += self.calc_1(i)
        if self.order >= 2:
            res += self.calc_2(i)
        return res

    def calc_0(self):
        return self.f0

    def calc_1(self, x):
        res = 0.
        num = 0
        for x1 in x:
            res += self.f1[num][x1]
            num += 1
        return res

    def calc_2(self, x):
        res = 0.
        num = 0
        for i1, x1 in enumerate(x[:-1]):
            for x2 in x[i1 + 1:]:
                res += self.f2[num][(x1, x2)]
                num += 1
        return res

    def cores(self, r=2, noise=1.E-10, only_near=False):
        if self.order < 1:
            raise ValueError('TT-cores may be constructed only if order >= 1')

        cores = self.cores_1(r, noise)

        if self.order >= 2:
            cores2_many = self.cores_2(r, only_near)
            cores = add_many([cores] + cores2_many, r=r)

        return cores

    def cores_1(self, r=2, noise=1.E-10):
        cores = []

        core = noise * np.random.randn(1, self.shapes[0], r)
        core[0, :, 0] = 1.
        core[0, :, 1] = self.f1_arr[0]
        cores.append(core)

        for i in range(1, self.d-1):
            core = noise * np.random.randn(r, self.shapes[i], r)
            core[0, :, 0] = 1.
            core[1, :, 1] = 1.
            core[0, :, 1] = self.f1_arr[i]
            cores.append(core)

        core = noise * np.random.randn(r, self.shapes[self.d-1], 1)
        core[0, :, 0] = self.f1_arr[self.d-1] + self.f0
        core[1, :, 0] = 1.
        cores.append(core)

        return cores

    def cores_2(self, r=2, only_near=False):
        mats = []
        num = 0
        for i1 in range(self.d-1):
            for i2 in ([i1+1] if only_near else range(i1+1, self.d)):
                shape = (self.shapes[i1], self.shapes[i2])
                mat = self.f2_arr[num].reshape(shape, order='C')
                mats.append(mat)
                num += 1

        cores = []
        num = 0
        for i1 in range(self.d-1):
            for i2 in ([i1+1] if only_near else range(i1+1, self.d)):
                cores.append(second_order_2_TT(mats[num], i1, i2, self.shapes))
                num += 1

        return cores


def anova(I_trn, Y_trn, r=2, order=1):
    """Build TT-tensor by TT-ANOVA from the given random tensor samples.

    Args:
        I_trn (np.ndarray): multi-indices for the tensor in the form of array
            of the shape [samples, d].
        Y_trn (np.ndarray): values of the tensor for multi-indices I in the form
            of array of the shape [samples].
        r (int): maximum rank of the constructed TT-tensor (should be > 0).
        order (int): order of the ANOVA decomposition (may be only 1 or 2).

    Returns:
        list: TT-tensor, which represents the TT-approximation for the tensor.

    Note:
        A class "ANOVA" that represents a wider set of methods for working with
        this decomposition is also available. See "teneva/core/anova.py" for
        more details. This function is just a wrapper for "ANOVA" class.

    """
    return ANOVA(I_trn, Y_trn, order).cores(r)


def core_one(n, r):
    return np.kron(np.ones([1, n, 1]), np.eye(r)[:, None, :])


def second_order_2_TT(A, i, j, shapes):
    if i > j:
        j, i = i, j
        A = A.T

    U, V = matrix_skeleton(A)
    r = U.shape[1]

    core1 = U.reshape(1, U.shape[0], r)
    core2 = V.reshape(r, V.shape[1], 1)

    cores = []
    for num, n in enumerate(shapes):
        if num < i or num > j:
            cores.append(np.ones([1, n, 1]))
        if num == i:
            cores.append(core1)
        if num == j:
            cores.append(core2)
        if i < num < j:
            cores.append(core_one(n, r))

    return cores



import os
import re
from setuptools import setup


def find_packages(package, basepath):
    packages = [package]
    for name in os.listdir(basepath):
        path = os.path.join(basepath, name)
        if not os.path.isdir(path):
            continue
        packages.extend(find_packages('%s.%s'%(package, name), path))
    return packages


here = os.path.abspath(os.path.dirname(__file__))
desc = 'Compact implementation of basic operations in the tensor-train (TT) format, including TT-SVD, TT-ALS, TT-ANOVA, TT-CROSS, TT-truncate for approximation of multidimensional arrays and multivariate functions'
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    desc_long = f.read()


with open(os.path.join(here, 'teneva/__init__.py'), encoding='utf-8') as f:
    text = f.read()
    version = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", text, re.M)
    version = version.group(1)


setup_args = dict(
    name='teneva',
    version=version,
    description=desc,
    long_description=desc_long,
    long_description_content_type='text/markdown',
    author='Andrei Chertkov',
    author_email='a.chertkov@skoltech.ru',
    url='https://github.com/AndreiChertkov/teneva',
    classifiers=[
        'Development Status :: 4 - Beta', # 3 - Alpha, 5 - Production/Stable
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Jupyter',
    ],
    keywords='low-rank representation tensor train format TT-decomposition cross approximation als anova',
    packages=find_packages('teneva', './teneva/'),
    python_requires='>=3.6',
    project_urls={
        'Source': 'https://github.com/AndreiChertkov/teneva',
    },
    license='MIT',
    license_files = ('LICENSE.txt',),
)


if __name__ == '__main__':
    setup(
        **setup_args,
        install_requires=['numpy', 'scipy', 'numba', 'matplotlib'],
        include_package_data=True)
